@@ -3,7 +3,6 @@ package pkg
 import (
 	"fmt"
 	"io"
-	"net/http"
 	"net/url"
 	"slices"
 	"strconv"
@@ -14,6 +13,9 @@ import (
 )
 
 var impactfulQueries []string
+var unkeyedQueries []string
+
+const NOOGPARAM = "NoOGParameter"
 
 func init() {
 }
@@ -22,55 +24,56 @@ func init() {
 func ScanCookies() reportResult {
 	var repResult reportResult
 	repResult.Technique = "Cookies"
-	for i, c := range Config.Website.Cookies {
-
+	i := 0
+	for k, v := range Config.Website.Cookies {
 		poison := "p" + randInt()
-		msg := fmt.Sprintf("Checking cookie %s (%d/%d)\n", c.Name, i+1, len(Config.Website.Cookies))
+		msg := fmt.Sprintf("Checking cookie %s (%d/%d)\n", k, i+1, len(Config.Website.Cookies))
 		Print(msg, NoColor)
+		i++
 
 		rUrl := Config.Website.Url.String()
 		cb := "cb" + randInt()
-		success := fmt.Sprintf("Cookie %s was successfully poisoned! cachebuster %s: %s poison: %s\n", c.Name, Config.Website.Cache.CBName, cb, poison)
-		identifier := c.Name + "=" + c.Value
-		msg = fmt.Sprintf("Overwriting %s=%s with %s=%s\n", c.Name, c.Value, c.Name, poison)
+		success := fmt.Sprintf("Cookie %s was successfully poisoned! cachebuster %s: %s poison: %s\n", k, Config.Website.Cache.CBName, cb, poison)
+		identifier := k + "=" + v
+		msg = fmt.Sprintf("Overwriting %s=%s with %s=%s\n", k, v, k, poison)
 		Print(msg, NoColor)
 
-		newCookie := *c
-		newCookie.Value = poison
+		newCookie := map[string]string{}
+		newCookie["key"] = k
+		newCookie["value"] = poison
 
 		rp := requestParams{
-			repResult:        &repResult,
-			headers:          []string{""},
-			values:           []string{""},
-			name:             c.Name,
-			identifier:       identifier,
-			poison:           poison,
-			url:              rUrl,
-			cb:               cb,
-			success:          success,
-			bodyString:       "",
-			forcePost:        false,
-			duplicateHeaders: false,
-			m:                nil,
-			newCookie:        newCookie,
+			repResult:  &repResult,
+			headers:    []string{""},
+			values:     []string{""},
+			name:       k,
+			identifier: identifier,
+			poison:     poison,
+			url:        rUrl,
+			cb:         cb,
+			success:    success,
+			bodyString: "",
+			forcePost:  false,
+			m:          nil,
+			newCookie:  newCookie,
 		}
-		responseSplittingHeader, _ := issueRequest(rp)
+		responseSplittingHeaders, _, _ := issueRequests(rp)
 
 		// check for response splitting, if poison was reflected in a header
-		if responseSplittingHeader != "" {
-			msg := fmt.Sprintf("Checking cookie %s for Response Splitting, because it was reflected in the header %s\n", c.Name, responseSplittingHeader)
+		for _, responseSplittingHeader := range responseSplittingHeaders {
+			msg := fmt.Sprintf("Checking cookie %s for Response Splitting, because it was reflected in the header %s\n", k, responseSplittingHeader)
 			PrintVerbose(msg, Cyan, 1)
 
 			rp.poison += getRespSplit()
 			rp.url = rUrl
 			rp.cb = "cb" + randInt()
-			rp.success = fmt.Sprintf("Cookie %s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", c.Name, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
+			rp.success = fmt.Sprintf("Cookie %s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", k, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
 			rp.identifier += " response splitting"
 
-			msg = fmt.Sprintf("Overwriting %s=%s with %s=%s\n", c.Name, c.Value, c.Name, rp.poison)
+			msg = fmt.Sprintf("Overwriting %s=%s with %s=%s\n", k, v, k, rp.poison)
 			Print(msg, NoColor)
 
-			issueRequest(rp)
+			issueRequests(rp)
 		}
 	}
 	return repResult
@@ -90,51 +93,60 @@ func ScanForwardingHeaders() reportResult {
 	}
 	port := strconv.Itoa(portInt)
 
-	values := []string{":" + port, ":@" + port, " " + port}
+	values := []string{Config.Website.Url.Host + ":" + port, Config.Website.Url.Host + ":@" + port, Config.Website.Url.Host + " " + port, Config.Website.Url.Host + ".p" + randInt(), "p" + randInt() + "." + Config.Website.Url.Host}
 	for _, value := range values {
 		PrintVerbose("Port-Number "+strconv.Itoa(portInt)+" was already present in websites response. Adding 1 to it.\n", NoColor, 2)
-		ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, false)
+		ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, NO_DUPE_HEADER)
 	}
 
 	// Duplicate Host header
-	header = "host"
-	poison := "p" + randInt()
-	values = []string{poison}
-	for _, value := range values {
-		ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, true)
+	headers := []string{"Host", "hOsT"}
+	dupes := []int{DUPE_HEADER_BEFORE, DUPE_HEADER_AFTER}
+	for _, header := range headers {
+		for _, dupe := range dupes {
+			poison := "p" + randInt()
+			ForwardHeadersTemplate(&repResult, []string{header}, []string{poison}, header, poison, dupe)
+		}
 	}
 
 	// X-Forwarded Headers
-	headers := []string{"X-Forwarded-Host", "X-Forwarded-Scheme"}
-	poison = "p" + randInt()
+	headers = []string{"X-Forwarded-Host", "X-Forwarded-Scheme"}
+	poison := "p" + randInt()
 	values = []string{poison, "nothttps"}
 	identifier := "X-Forwarded-Host and X-Forwarded-Scheme"
-	ForwardHeadersTemplate(&repResult, headers, values, identifier, poison, false)
+	ForwardHeadersTemplate(&repResult, headers, values, identifier, poison, NO_DUPE_HEADER)
 
 	values = []string{poison, "https"}
-	ForwardHeadersTemplate(&repResult, headers, values, identifier, poison, false)
+	ForwardHeadersTemplate(&repResult, headers, values, identifier, poison, NO_DUPE_HEADER)
 
 	// Forwarded Header
 	header = "Forwarded"
-	value := "host=" + "p" + randInt()
-	ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, false)
+	poison = "p" + randInt()
+	value := "host=" + poison
+	ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, poison, NO_DUPE_HEADER)
 
 	// X-Forwarded-Port Header
 	header = "X-Forwarded-Port"
 	value = port
-	ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, false)
+	ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, NO_DUPE_HEADER)
 
 	// X-Forwarded-Port Scheme
 	header = "X-Forwarded-Scheme"
 	values = []string{"http", "https"}
 	for _, value := range values {
-		ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, false)
+		ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, value, NO_DUPE_HEADER)
 	}
 
+	// X-Original-URL Header (https://samcurry.net/abusing-http-path-normalization-and-cache-poisoning-to-steal-rocket-league-accounts)
+	header = "X-Original-URL"
+	poison = "p" + randInt()
+	value = "http:\\\\" + poison + "/foo"
+	ForwardHeadersTemplate(&repResult, []string{header}, []string{value}, header, poison, NO_DUPE_HEADER)
+
 	return repResult
 }
 
-func ForwardHeadersTemplate(repResult *reportResult, headers []string, values []string, identifier string, poison string, duplicateHeaders bool) {
+func ForwardHeadersTemplate(repResult *reportResult, headers []string, values []string, identifier string, poison string, duplicateHeaders int) {
 	rUrl := Config.Website.Url.String()
 	cb := "cb" + randInt()
 	success := fmt.Sprintf("%s was successfully poisoned! cachebuster %s: %s poison: %s\n", headers, Config.Website.Cache.CBName, cb, values)
@@ -152,12 +164,11 @@ func ForwardHeadersTemplate(repResult *reportResult, headers []string, values []
 		forcePost:        false,
 		duplicateHeaders: duplicateHeaders,
 		m:                nil,
-		newCookie:        http.Cookie{},
 	}
-	responseSplittingHeader, _ := issueRequest(rp)
+	responseSplittingHeaders, _, _ := issueRequests(rp)
 
 	// check for response splitting, if poison was reflected in a header
-	if responseSplittingHeader != "" {
+	for _, responseSplittingHeader := range responseSplittingHeaders {
 		rp.values[0] += getRespSplit()
 		msg := fmt.Sprintf("Checking header(s) %s with value(s) %s for Response Splitting, because it was reflected in the header %s\n", rp.headers, rp.values, responseSplittingHeader)
 		PrintVerbose(msg, Cyan, 1)
@@ -168,7 +179,7 @@ func ForwardHeadersTemplate(repResult *reportResult, headers []string, values []
 		rp.success = fmt.Sprintf("%s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", headers, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.values)
 		rp.identifier += " response splitting"
 
-		issueRequest(rp)
+		issueRequests(rp)
 	}
 }
 
@@ -237,14 +248,12 @@ func ScanHeaders(headerList []string) reportResult {
 		header = strings.Trim(header, "\r")
 		if header == "" {
 			msg := fmt.Sprintf("Skipping empty header (%d/%d)\n", i+1, len(headerList))
-			PrintVerbose(msg, NoColor, 1)
+			PrintVerbose(msg, NoColor, 2)
 
 			wg.Done()
 			continue
 		}
 
-		header = http.CanonicalHeaderKey(header)
-
 		go func(i int, header string) {
 			defer wg.Done()
 			sem <- 1
@@ -259,25 +268,23 @@ func ScanHeaders(headerList []string) reportResult {
 			identifier := fmt.Sprintf("header %s", header)
 
 			rp := requestParams{
-				repResult:        &repResult,
-				headers:          []string{header},
-				values:           []string{poison},
-				name:             header,
-				identifier:       identifier,
-				poison:           poison,
-				url:              rUrl,
-				cb:               cb,
-				success:          success,
-				bodyString:       "",
-				forcePost:        false,
-				duplicateHeaders: false,
-				m:                &m,
-				newCookie:        http.Cookie{},
+				repResult:  &repResult,
+				headers:    []string{header},
+				values:     []string{poison},
+				name:       header,
+				identifier: identifier,
+				poison:     poison,
+				url:        rUrl,
+				cb:         cb,
+				success:    success,
+				bodyString: "",
+				forcePost:  false,
+				m:          &m,
 			}
-			responseSplittingHeader, _ := issueRequest(rp)
+			responseSplittingHeaders, _, _ := issueRequests(rp)
 
 			// check for response splitting, if poison was reflected in a header
-			if responseSplittingHeader != "" {
+			for _, responseSplittingHeader := range responseSplittingHeaders {
 				msg := fmt.Sprintf("Testing now (%d/%d) %s for Response Splitting, because it was reflected in the header %s\n", i+1, len(headerList), header, responseSplittingHeader)
 				PrintVerbose(msg, Cyan, 1)
 
@@ -287,7 +294,7 @@ func ScanHeaders(headerList []string) reportResult {
 				rp.success = fmt.Sprintf("Header %s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", header, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
 				rp.identifier += " response splitting"
 
-				issueRequest(rp)
+				issueRequests(rp)
 			}
 		}(i, header)
 
@@ -302,12 +309,12 @@ func ScanParameters(parameterList []string) reportResult {
 	var repResult reportResult
 	repResult.Technique = "Parameters"
 
-	var parametersToTest []string
-
+	parametersToTest := parameterList
 	for k := range Config.Website.Queries {
-		parametersToTest = append(parametersToTest, k) // add custom parameters to list
+		if !slices.Contains(parameterList, k) { // only add parameters which are not already in the list
+			parametersToTest = append(parametersToTest, k) // add custom parameters to list
+		}
 	}
-	parametersToTest = append(parametersToTest, parameterList...)
 
 	threads := Config.Threads
 	if Config.Website.Cache.CBisHTTPMethod {
@@ -320,6 +327,7 @@ func ScanParameters(parameterList []string) reportResult {
 	var m sync.Mutex
 
 	impactfulQueries = []string{}
+	unkeyedQueries = []string{}
 
 	msg := fmt.Sprintf("Testing %d parameters\n", len(parametersToTest))
 	PrintVerbose(msg, NoColor, 1)
@@ -347,34 +355,38 @@ func ScanParameters(parameterList []string) reportResult {
 			success := fmt.Sprintf("Query Parameter %s was successfully poisoned! cachebuster %s: %s poison: %s\n", parameter, Config.Website.Cache.CBName, cb, poison)
 			identifier := fmt.Sprintf("parameter %s", parameter)
 
+			ogValue := NOOGPARAM
 			if strings.Contains(strings.ToLower(rUrl), "?"+parameter+"=") || strings.Contains(strings.ToLower(rUrl), "&"+parameter+"=") { // remove param if it already existed, so that it will be set only one time and that being with the poison value
-				rUrl, _ = removeParam(rUrl, parameter)
+				rUrl, ogValue, _ = removeParam(rUrl, parameter)
 			}
 
 			rp := requestParams{
-				repResult:        &repResult,
-				headers:          []string{""},
-				values:           []string{poison},
-				parameters:       []string{parameter + "=" + poison},
-				name:             parameter,
-				identifier:       identifier,
-				poison:           poison,
-				url:              rUrl,
-				cb:               cb,
-				success:          success,
-				bodyString:       "",
-				forcePost:        false,
-				duplicateHeaders: false,
-				newCookie:        http.Cookie{},
-				m:                &m,
+				repResult:  &repResult,
+				headers:    []string{""},
+				values:     []string{poison},
+				parameters: []string{parameter + "=" + poison},
+				name:       parameter,
+				identifier: identifier,
+				poison:     poison,
+				ogParam:    parameter + "=" + ogValue,
+				url:        rUrl,
+				cb:         cb,
+				technique:  "parameter",
+				success:    success,
+				bodyString: "",
+				forcePost:  false,
+				m:          &m,
 			}
-			responseSplittingHeader, appendParameter := issueRequest(rp)
+			responseSplittingHeaders, impactful, unkeyed := issueRequests(rp)
 
-			if appendParameter {
+			if impactful && !unkeyed {
 				impactfulQueries = append(impactfulQueries, parameter)
+			} else if unkeyed {
+				unkeyedQueries = append(unkeyedQueries, parameter)
 			}
+
 			// check for response splitting, if poison was reflected in a header
-			if responseSplittingHeader != "" {
+			for _, responseSplittingHeader := range responseSplittingHeaders {
 				msg := fmt.Sprintf("Testing now Parameter (%d/%d) %s for Response Splitting, because it was reflected in the header %s\n", i+1, len(parametersToTest), parameter, responseSplittingHeader)
 				PrintVerbose(msg, Cyan, 1)
 
@@ -384,7 +396,7 @@ func ScanParameters(parameterList []string) reportResult {
 				rp.cb = "cb" + randInt()
 				rp.success = fmt.Sprintf("Query Parameter %s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", parameter, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
 				rp.identifier += " response splitting"
-				issueRequest(rp)
+				issueRequests(rp)
 			}
 		}(i, parameter)
 
@@ -461,24 +473,22 @@ func ScanFatGET() reportResult {
 				success := fmt.Sprintf("Query Parameter %s was successfully poisoned via %s! cachebuster %s: %s poison:%s\n", s, identifier, Config.Website.Cache.CBName, cb, poison)
 
 				rp := requestParams{
-					repResult:        &repResult,
-					headers:          []string{headers[method]},
-					values:           []string{values[method]},
-					identifier:       identifier,
-					poison:           poison,
-					url:              rUrl,
-					cb:               cb,
-					success:          success,
-					bodyString:       bodyString,
-					forcePost:        forcePost,
-					duplicateHeaders: false,
-					m:                &m,
-					newCookie:        http.Cookie{},
+					repResult:  &repResult,
+					headers:    []string{headers[method]},
+					values:     []string{values[method]},
+					identifier: identifier,
+					poison:     poison,
+					url:        rUrl,
+					cb:         cb,
+					success:    success,
+					bodyString: bodyString,
+					forcePost:  forcePost,
+					m:          &m,
 				}
-				responseSplittingHeader, _ := issueRequest(rp)
+				responseSplittingHeaders, _, _ := issueRequests(rp)
 
 				// check for response splitting, if poison was reflected in a header
-				if responseSplittingHeader != "" {
+				for _, responseSplittingHeader := range responseSplittingHeaders {
 					msg := fmt.Sprintf("Testing now (%d/%d) %s for Response Splitting, because it was reflected in the header %s\n", i+1, len(impactfulQueries), s, responseSplittingHeader)
 					PrintVerbose(msg, Cyan, 1)
 
@@ -489,7 +499,7 @@ func ScanFatGET() reportResult {
 					rp.identifier += " response splitting"
 					rp.success = fmt.Sprintf("Query Parameter %s successfully poisoned the header %s via %s with Response Splitting! cachebuster %s: %s poison:%s\n", s, responseSplittingHeader, identifier, Config.Website.Cache.CBName, rp.cb, rp.poison)
 
-					issueRequest(rp)
+					issueRequests(rp)
 				}
 			}(i, s, poison)
 		}
@@ -516,11 +526,12 @@ func ScanParameterCloaking() reportResult {
 		Print(msg, Cyan)
 	}
 
-	utm_parameters := []string{"utm_source", "utm_medium", "utm_campaign", "utm_content", "utm_term"}
-	unkeyed_parameters := []string{}
+	utm_parameters := []string{"utm_source", "utm_medium", "utm_campaign", "utm_content", "utm_term", "gad_campaignid"}
 	parameters_to_test := utm_parameters
-	for k := range Config.Website.Queries {
-		parameters_to_test = append(parameters_to_test, k)
+	for _, k := range unkeyedQueries {
+		if !slices.Contains(parameters_to_test, k) { // only add parameters which are not already in the list
+			parameters_to_test = append(parameters_to_test, k)
+		}
 	}
 
 	/***********TODO Check if urlWithCb already contains utm parameter.
@@ -546,6 +557,8 @@ func ScanParameterCloaking() reportResult {
 	sem := make(chan int, threads)
 	var wg sync.WaitGroup
 	var m sync.Mutex
+
+	unkeyed_parameters := []string{}
 	cache := Config.Website.Cache
 	if cache.Indicator == "" || cache.TimeIndicator {
 		msg := "hit/miss isn't verbose. Can't check which parameters unkeyed, so all utm_parameters and query parameters will be used\n"
@@ -581,8 +594,12 @@ func ScanParameterCloaking() reportResult {
 					m.Unlock()
 					return
 				}
-				indicValue := respHeader.Get(cache.Indicator) // TODO add check for timebased cache indicator! + remove then cache.TimeIndicator from the if above
-				if checkCacheHit(indicValue, cache.Indicator) {
+				hit := false
+				for _, v := range respHeader[cache.Indicator] {
+					indicValue := strings.TrimSpace(strings.ToLower(v))
+					hit = hit || checkCacheHit(indicValue, cache.Indicator)
+				} // TODO add check for timebased cache indicator! + remove then cache.TimeIndicator from the if above
+				if hit {
 					m.Lock()
 					unkeyed_parameters = append(unkeyed_parameters, s)
 					m.Unlock()
@@ -654,26 +671,25 @@ func ScanParameterCloaking() reportResult {
 				}
 
 				rp := requestParams{
-					repResult:        &repResult,
-					headers:          []string{""},
-					values:           []string{poison},
-					parameters:       []string{u + "=foobar" + cloak + s + "=" + poison},
-					identifier:       identifier,
-					poison:           poison,
-					url:              rUrl,
-					cb:               cb,
-					prependCB:        prependCB,
-					success:          success,
-					bodyString:       "",
-					forcePost:        false,
-					duplicateHeaders: false,
-					m:                &m,
-					newCookie:        http.Cookie{},
+					repResult:  &repResult,
+					headers:    []string{""},
+					values:     []string{poison},
+					parameters: []string{u + "=foobar" + cloak + s + "=" + poison},
+					identifier: identifier,
+					poison:     poison,
+					url:        rUrl,
+					cb:         cb,
+					prependCB:  prependCB,
+					success:    success,
+					bodyString: "",
+					forcePost:  false,
+					m:          &m,
+					newCookie:  nil,
 				}
-				responseSplittingHeader, _ := issueRequest(rp)
+				responseSplittingHeaders, _, _ := issueRequests(rp)
 
 				// check for response splitting, if poison was reflected in a header
-				if responseSplittingHeader != "" {
+				for _, responseSplittingHeader := range responseSplittingHeaders {
 					msg := fmt.Sprintf("Testing now Parameter Cloaking (%d/%d) %s%s%s for Response Splitting, because it was reflected in the header %s\n", iu+is+1, len(impactfulQueries)*len(unkeyed_parameters), u, cloak, s, responseSplittingHeader)
 					PrintVerbose(msg, Cyan, 1)
 
@@ -684,7 +700,7 @@ func ScanParameterCloaking() reportResult {
 					rp.success = fmt.Sprintf("Query Parameter %s successfully poisoned the header %s with Response Splitting using %s with Parameter Cloaking! cachebuster %s:%s poison:%s\n", s, responseSplittingHeader, u, Config.Website.Cache.CBName, rp.cb, rp.poison)
 					rp.identifier += " response splitting"
 
-					issueRequest(rp)
+					issueRequests(rp)
 				}
 			}(iu, u, is, s, poison)
 		}
@@ -740,15 +756,16 @@ func ScanParameterPollution() reportResult {
 			defer func() { <-sem }() // Freigabe der Semaphore, egal was passiert. Dadurch werden Deadlocks verhindert
 
 			url := Config.Website.Url.String()
+			ogValue := "foobar"
 			if strings.Contains(strings.ToLower(url), "?"+s+"=") || strings.Contains(strings.ToLower(url), "&"+s+"=") {
-				url, _ = removeParam(url, s)
+				url, ogValue, _ = removeParam(url, s)
 			}
 
 			var parameters []string
 			if is >= len(impactfulQueries)/2 {
-				parameters = []string{s + "=" + poison, s + "=foobar"}
+				parameters = []string{s + "=" + poison, s + "=" + ogValue}
 			} else {
-				parameters = []string{s + "=foobar", s + "=" + poison}
+				parameters = []string{s + "=" + ogValue, s + "=" + poison}
 			}
 
 			msg := fmt.Sprintf("Testing now Parameter Pollution (%d/%d) %s\n", is+1, len(impactfulQueries)*2, s)
@@ -758,27 +775,27 @@ func ScanParameterPollution() reportResult {
 			identifier := fmt.Sprintf("parameter Pollution %s", s)
 
 			rp := requestParams{
-				repResult:        &repResult,
-				headers:          []string{""},
-				values:           []string{poison},
-				parameters:       parameters,
-				technique:        "pollution",
-				identifier:       identifier,
-				poison:           poison,
-				url:              url,
-				cb:               cb,
-				prependCB:        prependCB,
-				success:          success,
-				bodyString:       "",
-				forcePost:        false,
-				duplicateHeaders: false,
-				m:                &m,
-				newCookie:        http.Cookie{},
+				repResult:  &repResult,
+				headers:    []string{""},
+				values:     []string{poison},
+				parameters: parameters,
+				technique:  "pollution",
+				identifier: identifier,
+				poison:     poison,
+				url:        url,
+				cb:         cb,
+				ogParam:    s + "=" + ogValue,
+				prependCB:  prependCB,
+				success:    success,
+				bodyString: "",
+				forcePost:  false,
+				m:          &m,
+				newCookie:  nil,
 			}
-			responseSplittingHeader, _ := issueRequest(rp)
+			responseSplittingHeaders, _, _ := issueRequests(rp)
 
 			// check for response splitting, if poison was reflected in a header
-			if responseSplittingHeader != "" {
+			for _, responseSplittingHeader := range responseSplittingHeaders {
 				msg := fmt.Sprintf("Testing now Parameter Pollution (%d/%d) %s for Response Splitting, because it was reflected in the header %s\n", is+1, len(impactfulQueries)*2, s, responseSplittingHeader)
 				PrintVerbose(msg, Cyan, 1)
 
@@ -796,7 +813,119 @@ func ScanParameterPollution() reportResult {
 				rp.success = fmt.Sprintf("Query Parameter %s successfully poisoned the header %s with Response Splitting with Parameter Pollution! cachebuster %s:%s poison:%s\n", s, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
 				rp.identifier += " response splitting"
 
-				issueRequest(rp)
+				issueRequests(rp)
+			}
+		}(is, s, poison, is >= len(impactfulQueries)/2)
+	}
+	wg.Wait()
+
+	return repResult
+}
+
+/* Check for Parameter Encoding */
+func ScanParameterEncoding() reportResult {
+	var repResult reportResult
+	repResult.Technique = "Parameter Encoding"
+
+	if len(impactfulQueries) == 0 {
+		msg := "No impactful query parameters were found beforehand. Run the query parameter scan (maybe with a different wordlist)."
+		Print(msg+"\n", Yellow)
+		repResult.HasError = true
+		repResult.ErrorMessages = append(repResult.ErrorMessages, msg)
+		return repResult
+	} else {
+		msg := fmt.Sprintf("The following parameters were found to be impactful and will be tested for parameter encoding: %s\n", impactfulQueries)
+		Print(msg, Cyan)
+	}
+
+	threads := Config.Threads
+	if Config.Website.Cache.CBisHTTPMethod {
+		threads = 1 // No multithreading if HTTP Method is used... Otherwise there will be a lot of false negatives/positives
+		PrintVerbose("Can only scan single threaded because a HTTP Method is used as Cachebuster...\n", Yellow, 1)
+	}
+	sem := make(chan int, threads)
+	var wg sync.WaitGroup
+	var m sync.Mutex
+
+	if len(impactfulQueries) > 500 { // only test first 500 impactful queries. TODO decrease amount as such a high amount most likely means they are false positives anyways
+		shortenedQueries := []string{}
+		for i := range 500 {
+			shortenedQueries = append(shortenedQueries, impactfulQueries[i])
+		}
+		impactfulQueries = shortenedQueries
+	}
+	impactfulQueries = append(impactfulQueries, impactfulQueries...) // we want to test each impactful query parameter 2times. One time with the poison in the first and second with the poison in the second appearance
+
+	wg.Add(len(impactfulQueries))
+
+	for is, s := range impactfulQueries {
+		poison := "p" + randInt()
+		prependCB := false // shall be true for the second test of the cachebuster
+
+		go func(is int, s string, poison string, secondHalf bool) {
+			defer wg.Done()
+			sem <- 1
+			defer func() { <-sem }() // Freigabe der Semaphore, egal was passiert. Dadurch werden Deadlocks verhindert
+
+			url := Config.Website.Url.String()
+			ogValue := "foobar"
+			if strings.Contains(strings.ToLower(url), "?"+s+"=") || strings.Contains(strings.ToLower(url), "&"+s+"=") {
+				url, ogValue, _ = removeParam(url, s)
+			}
+
+			var parameters []string
+			if is >= len(impactfulQueries)/2 {
+				parameters = []string{urlEncodeAll(s) + "=" + poison, s + "=" + ogValue}
+			} else {
+				parameters = []string{s + "=" + ogValue, urlEncodeAll(s) + "=" + poison}
+			}
+
+			msg := fmt.Sprintf("Testing now Parameter Encoding (%d/%d) %s\n", is+1, len(impactfulQueries)*2, s)
+			PrintVerbose(msg, NoColor, 2)
+			cb := "cb" + randInt()
+			success := fmt.Sprintf("Query Parameter %s was successfully poisoned via Parameter Encoding! cachebuster %s:%s poison:%s\n", s, Config.Website.Cache.CBName, cb, poison)
+			identifier := fmt.Sprintf("parameter Encoding %s", s)
+
+			rp := requestParams{
+				repResult:  &repResult,
+				headers:    []string{""},
+				values:     []string{poison},
+				parameters: parameters,
+				technique:  "encoding",
+				identifier: identifier,
+				poison:     poison,
+				url:        url,
+				cb:         cb,
+				ogParam:    s + "=" + ogValue,
+				prependCB:  prependCB,
+				success:    success,
+				bodyString: "",
+				forcePost:  false,
+				m:          &m,
+				newCookie:  nil,
+			}
+			responseSplittingHeaders, _, _ := issueRequests(rp)
+
+			// check for response splitting, if poison was reflected in a header
+			for _, responseSplittingHeader := range responseSplittingHeaders {
+				msg := fmt.Sprintf("Testing now Parameter Encoding (%d/%d) %s for Response Splitting, because it was reflected in the header %s\n", is+1, len(impactfulQueries)*2, s, responseSplittingHeader)
+				PrintVerbose(msg, Cyan, 1)
+
+				rp.url = url
+				rp.cb = "cb" + randInt()
+				rp.poison += getRespSplit()
+
+				if is >= len(impactfulQueries) {
+					parameters = []string{urlEncodeAll(s) + "=" + rp.poison, s + "=foobar"}
+				} else {
+					parameters = []string{s + "=foobar", urlEncodeAll(s) + "=" + rp.poison}
+				}
+
+				rp.parameters = parameters
+				rp.success = fmt.Sprintf("Query Parameter %s successfully poisoned the header %s with Response Splitting with Parameter Encoding! cachebuster %s:%s poison:%s\n", s, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.poison)
+				rp.identifier += " response splitting"
+
+				issueRequests(rp)
 			}
 		}(is, s, poison, is >= len(impactfulQueries)/2)
 	}
@@ -810,19 +939,23 @@ func DOS() reportResult {
 	var repResult reportResult
 	repResult.Technique = "DOS"
 
-	// TODO: Ist nur Header Value oder auch Header Name ausschlaggebend?
 	hho(&repResult)
 
 	// HMC (Header Metachar Character)
-	// TODO: Check for more META CHARACTERS?
-	//TODO: Change to other header, which is probably whitelisted
 	headers := []string{"X-Metachar-Header"}
-	values := []string{"\\n", "\\r", "\\a", "\\0", "\\b", "\\e", "\\v", "\\f", "\\u0000"}
+	values := []string{"n\nn", "r\rr", "a\aa", "x00\x00x00", "b\bb", "x1b\x1bx1b", "v\vv", "f\ff", "u0000\u0000u0000"} // TODO put not functional meta chars, due to fasthttp header value restrictions, in the value using \r\n in the header name
 
 	for _, header := range headers {
 		headerDOSTemplate(&repResult, values, header, "HMC ", true)
 	}
 
+	headers = []string{"X-Meta\nchar-Header", "X-Meta\rchar-Header", "X-Meta\achar-Header", "X-Meta\x00char-Header", "X-Meta\bchar-Header", "X-Meta\x1bchar-Header", "X-Meta\vchar-Header", "X-Meta\fchar-Header", "X-Meta\u0000char-Header"}
+	values = []string{"n", "r", "a", "x00", "b", "x1b", "v", "ff", "u0000"}
+
+	for i, header := range headers {
+		headerDOSTemplate(&repResult, []string{values[i]}, header, "HMC ", true)
+	}
+
 	// HMO (HTTP Method Override)
 	values = []string{"GET", "POST", "DELETE", "HEAD", "OPTIONS", "CONNECT", "PATCH", "PUT", "TRACE", "NONSENSE"}
 	headers = []string{"X-HTTP-Method-Override", "X-HTTP-Method", "X-Method-Override"}
@@ -850,7 +983,7 @@ func DOS() reportResult {
 
 	// DOS via Max-Forwards (Webserver/Cache returns request)
 	values = []string{"0", "1", "2"}
-	headerDOSTemplate(&repResult, values, "Max-Forwards", "max-forwards ", true)
+	headerDOSTemplate(&repResult, values, "Max-Forwards", "", true)
 
 	// DOS via waf blocking because of a blacklist word
 	// TODO: change header to probably whitelisted header, More Blacklist words?
@@ -858,25 +991,37 @@ func DOS() reportResult {
 	headerDOSTemplate(&repResult, values, "Any-Header", "blacklist ", true)
 
 	// DOS via Range
-	values = []string{"bytes=cow"}
+	values = []string{"bytes=m10x", "bytes=9-4", "bytes=-1024,0", "bytes=0-,0-,0-,0-"}
 	headerDOSTemplate(&repResult, values, "Range", "", true)
 
 	// DOS via X-Forwarded-Protocol
-	values = []string{"http", "https", "ssl", "nonsense"}
+	values = []string{"http", "https", "ssl", "m10x"}
 	headerDOSTemplate(&repResult, values, "X-Forwarded-Protocol", "", true)
 
-	// DOS via X-Forwarded-Protocol
-	values = []string{"http", "https", "nothttps", "nonsense"}
+	// DOS via X-Forwarded-Scheme
+	values = []string{"http", "https", "nothttps", "m10x"}
 	headerDOSTemplate(&repResult, values, "X-Forwarded-Scheme", "", true)
 
 	// DOS via X-Fordwarded-SSL
-	values = []string{"on", "off", "nonsense"}
+	values = []string{"on", "off", "m10x"}
 	headerDOSTemplate(&repResult, values, "X-Forwarded-SSL", "", true)
 
+	// DOS via Upgrade
+	values = []string{"HTTP/0.9", "Websocket, RTA/x11", "HTTP/2.0, SHTTP/1.3, IRC/6.9", "m10x"}
+	headerDOSTemplate(&repResult, values, "Upgrade", "", true)
+
 	// DOS via invalid Content-Type
 	values = []string{"m10x"}
 	headerDOSTemplate(&repResult, values, "Content-Type", "", true)
 
+	// DOS via middleware prefetch (next.js specific, CVE-2023-46298) TODO check for {} in response
+	values = []string{"1"}
+	headerDOSTemplate(&repResult, values, "X-Middleware-Prefetch", "", true)
+
+	// DOS via Rsc (next.js specific) TODO check for format in response
+	values = []string{"1"}
+	headerDOSTemplate(&repResult, values, "Rsc", "", true)
+
 	return repResult
 }
 
@@ -910,7 +1055,7 @@ func hho(repResult *reportResult) {
 			headers := []string{}
 			values := []string{}
 
-			for i := 0; i < repetition; i++ {
+			for i := range repetition {
 				headername := fmt.Sprintf("X-Oversized-Header-%d", i+1)
 				value := "Big-Value-000000000000000000000000000000000000000000000000000000000000000000000000000000"
 				headers = append(headers, headername)
@@ -919,41 +1064,20 @@ func hho(repResult *reportResult) {
 
 			rUrl := Config.Website.Url.String()
 			cb := "cb" + randInt()
+			success := fmt.Sprintf("HHO DOS was successfully poisoned! cachebuster %s: %s \n%s\n", Config.Website.Cache.CBName, cb, rUrl)
 			identifier := fmt.Sprintf("HHO with limit of %dk bytes", limit)
 			rp := requestParams{
+				repResult:  repResult,
 				headers:    headers,
 				values:     values,
 				identifier: identifier,
 				url:        rUrl,
 				cb:         cb,
-			}
-			_, statusCode1, request, _, err := firstRequest(rp)
-			if err != nil {
-				if err.Error() != "stop" {
-					m.Lock()
-					repResult.HasError = true
-					repResult.ErrorMessages = append(repResult.ErrorMessages, err.Error())
-					m.Unlock()
-				}
-				return
+				success:    success,
+				m:          &m,
 			}
 
-			// send second request also with cb
-			_, statusCode2, respHeader, err := secondRequest(rp)
-			if err != nil {
-				if err.Error() != "stop" {
-					m.Lock()
-					repResult.HasError = true
-					repResult.ErrorMessages = append(repResult.ErrorMessages, err.Error())
-					m.Unlock()
-				}
-				return
-			}
-
-			msg = fmt.Sprintf("HHO DOS was successfully poisoned! cachebuster %s: %s \n%s\n", Config.Website.Cache.CBName, cb, request.URL)
-			m.Lock()
-			_ = checkPoisoningIndicators(repResult, request, msg, "", "", statusCode1, statusCode2, false, respHeader, false)
-			m.Unlock()
+			_, _, _ = issueRequests(rp)
 		}(repetition)
 	}
 
@@ -981,32 +1105,31 @@ func headerDOSTemplate(repResult *reportResult, values []string, header string,
 			sem <- 1
 			defer func() { <-sem }() // Freigabe der Semaphore, egal was passiert. Dadurch werden Deadlocks verhindert
 
-			msg := fmt.Sprintf("Testing now %s Header DOS with %s\n", header, value)
+			msg := fmt.Sprintf("Testing now %q Header DOS with %q\n", header, value) // %q for raw printing of control characters
 			PrintVerbose(msg, NoColor, 2)
 			rUrl := Config.Website.Url.String()
 			cb := "cb" + randInt()
-			success := fmt.Sprintf("%sDOS with header %s was successfully poisoned! cachebuster %s: %s poison: %s\n", msgextra, header, Config.Website.Cache.CBName, cb, value)
-			identifier := fmt.Sprintf("%s%s with %s", msgextra, header, value)
+			success := fmt.Sprintf("%sDOS with header %q was successfully poisoned! cachebuster %s: %s poison: %q\n", msgextra, header, Config.Website.Cache.CBName, cb, value)
+			identifier := fmt.Sprintf("%s%q with %q", msgextra, header, value)
 
 			rp := requestParams{
-				repResult:        repResult,
-				headers:          []string{header},
-				values:           []string{value},
-				identifier:       identifier,
-				poison:           "",
-				url:              rUrl,
-				cb:               cb,
-				success:          success,
-				bodyString:       "",
-				forcePost:        false,
-				duplicateHeaders: false,
-				m:                &m,
-				newCookie:        http.Cookie{},
+				repResult:  repResult,
+				headers:    []string{header},
+				values:     []string{value},
+				identifier: identifier,
+				poison:     "",
+				url:        rUrl,
+				cb:         cb,
+				success:    success,
+				bodyString: "",
+				forcePost:  false,
+				m:          &m,
+				newCookie:  nil,
 			}
-			responseSplittingHeader, _ := issueRequest(rp)
+			responseSplittingHeaders, _, _ := issueRequests(rp)
 
 			// check for response splitting, if poison was reflected in a header
-			if responseSplittingHeader != "" {
+			for _, responseSplittingHeader := range responseSplittingHeaders {
 				msg := fmt.Sprintf("Testing now %s Header DOS with %s\n for Response Splitting, because it was reflected in the header %s", header, value, responseSplittingHeader)
 				PrintVerbose(msg, Cyan, 1)
 
@@ -1016,7 +1139,7 @@ func headerDOSTemplate(repResult *reportResult, values []string, header string,
 				rp.success = fmt.Sprintf("%sDOS with header %s successfully poisoned the header %s with Response Splitting! cachebuster %s: %s poison: %s\n", msgextra, header, responseSplittingHeader, Config.Website.Cache.CBName, rp.cb, rp.values[0])
 				rp.identifier += getRespSplit() + " with response splitting"
 
-				issueRequest(rp)
+				issueRequests(rp)
 			}
 		}(value, httpConform)
 	}
@@ -1027,7 +1150,14 @@ func ScanCSS() reportResult {
 	var repResult reportResult
 	repResult.Technique = "CSS poisoning"
 
-	bodyReader := strings.NewReader(Config.Website.Body)
+	webStruct, err := GetWebsite(Config.Website.Url.String(), false, false) // get body without cachebuster. TODO use response w/o cachebuster from recon, so it doesn't have to be fetched again
+	if err != nil {
+		msg := fmt.Sprintf("Error while fetching css files %s: %s\n", Config.Website.Url.String(), err.Error())
+		Print(msg, Red)
+		repResult.ErrorMessages = append(repResult.ErrorMessages, msg)
+		return repResult
+	}
+	bodyReader := strings.NewReader(webStruct.Body) // use body without cachebuster, so the css files can be found
 	tokenizer := html.NewTokenizer(bodyReader)
 
 	var urls []string
@@ -1097,6 +1227,9 @@ func ScanCSS() reportResult {
 			//Print(msg, NoColor)
 
 			urlWithCb, cb := addCachebusterParameter(url, "", Config.Website.Cache.CBName, false)
+			var repCheck reportCheck
+			repCheck.URL = url
+			repCheck.Identifier = "n/a"
 
 			identifier := url
 			rp := requestParams{
@@ -1104,7 +1237,7 @@ func ScanCSS() reportResult {
 				url:        urlWithCb,
 				cb:         "cb" + randInt(),
 			}
-			body, _, request, _, err := firstRequest(rp)
+			body, status, repRequest, _, err := firstRequest(rp)
 			if err != nil {
 				if err.Error() != "stop" {
 					m.Lock()
@@ -1114,13 +1247,17 @@ func ScanCSS() reportResult {
 				}
 				return
 			}
+			repCheck.Request = repRequest
+			if status != 200 {
+				return
+			}
 
 			if strings.Contains(string(body), cb) {
 				msg = fmt.Sprintf("The following CSS file reflects the url with the cb %s\n%s\n", cb, url)
-				Print(msg, Green)
+				Print(msg, Cyan)
 			}
 
-			body, _, _, err = secondRequest(rp)
+			body, _, repRequest, _, err = secondRequest(rp)
 			if err != nil {
 				if err.Error() != "stop" {
 					m.Lock()
@@ -1130,17 +1267,20 @@ func ScanCSS() reportResult {
 				}
 				return
 			}
+			repCheck.SecondRequest = &repRequest
 
 			if strings.Contains(string(body), cb) {
 				PrintNewLine()
-				msg = fmt.Sprintf("A CSS file was successfully poisoned! cachebuster %s: %s\nURL: %s\n", Config.Website.Cache.CBName, cb, request.URL)
+				msg = fmt.Sprintf("A CSS file was successfully poisoned! cachebuster %s: %s\nURL: %s\n", Config.Website.Cache.CBName, cb, url)
 				Print(msg, Green)
-				msg = "Reason: CSS reflects URL\n"
+				Reason := "CSS reflects URL"
+				msg = fmt.Sprintf("Reason: %s\n", Reason)
 				Print(msg, Green)
+				repCheck.Reason = Reason
 
 				m.Lock()
 				repResult.Vulnerable = true
-				repResult.Requests = append(repResult.Requests, request)
+				repResult.Checks = append(repResult.Checks, repCheck)
 				m.Unlock()
 			}
 		}(url)
