@@ -1,8 +1,6 @@
 package main
 
 import (
-	"crypto/tls"
-	"crypto/x509"
 	"fmt"
 	"log"
 	"net/http"
@@ -12,13 +10,13 @@ import (
 	"time"
 
 	"github.com/Hackmanit/Web-Cache-Vulnerability-Scanner/pkg"
-	"golang.org/x/net/http2"
 )
 
-const version = "1.4.2"
+const version = "2.0.0"
 
 var (
 	currentDate      string
+	filePath         string
 	report           pkg.Report
 	completedFile    *os.File
 	proxyURL         *url.URL
@@ -32,7 +30,6 @@ var (
 func main() {
 	//pkg.ReadConfigFile()
 	pkg.ParseFlags(version)
-	pkg.Statistics = map[string]int{}
 	/*****************************/
 
 	/**** SET EXPORT STRUCT ****/
@@ -61,9 +58,11 @@ func main() {
 		}
 	}
 
+	filePath = fmt.Sprintf("%sWCVS_%s_%s", pkg.Config.GeneratePath, currentDate, pkg.RandomString(8))
+
 	/* Setting Logoutput to Log file */
-	if !pkg.Config.NoLog {
-		f, err := os.OpenFile(pkg.Config.GeneratePath+currentDate+"_WCVS_Log", os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
+	if pkg.Config.GenerateLog {
+		f, err := os.OpenFile(filePath+"_Log.txt", os.O_WRONLY|os.O_CREATE|os.O_APPEND, 0644)
 		if err != nil {
 			msg := fmt.Sprintf("Error while creating/opening Log File: %s\n", err.Error())
 			pkg.PrintFatal(msg)
@@ -106,17 +105,14 @@ func main() {
 	}
 
 	if pkg.Config.GenerateReport {
-		pkg.GenerateReport(report, currentDate)
+		pkg.GenerateReport(report, filePath)
 	}
 	if pkg.Config.GenerateCompleted {
 		completedFile = createCompletedURLs()
 	}
+	pkg.InitClient()
 	/***************************/
 
-	/* Setting up proxy (e.g. burp), if wanted */
-	proxyURL = setProxy()
-	/*******************************************/
-
 	// Reading header wordlist, only if it is needed
 	if noTestPreference || strings.Contains(pkg.Config.OnlyTest, "header") || (pkg.Config.SkipTest != "" && !strings.Contains(pkg.Config.SkipTest, "header")) {
 		if pkg.Config.HeaderWordlist != "" {
@@ -218,25 +214,33 @@ func main() {
 
 	if pkg.Config.GenerateReport {
 		report.Duration = duration.String()
-		pkg.GenerateReport(report, currentDate)
+		pkg.GenerateReport(report, filePath)
 	}
 }
 
 func runTests(rec int, u string, progress string, foundUrls *[]string, stat string) {
 	var repWebsite pkg.ReportWebsite
 	var err error
-	repWebsite.URL = u
 
 	msg := fmt.Sprintf("\nTesting website%s: %s\n", progress, u)
 	pkg.Print(msg, pkg.NoColor)
 	pkg.Print("===============================================================\n", pkg.NoColor)
 
+	if !strings.HasPrefix(u, "http://") && !strings.HasPrefix(u, "https://") {
+		if pkg.Config.UseHTTP {
+			u = "http://" + u
+		} else {
+			u = "https://" + u
+		}
+	}
+	repWebsite.URL = u
+
 	/* Setting up client: cookies and noredirect */
 	msg = "Setting up client\n"
 	pkg.PrintVerbose(msg, pkg.NoColor, 2)
 
 	// Setting cookies, specified by setcookies
-	pkg.Config.Website.Cookies = []*http.Cookie{}
+	pkg.Config.Website.Cookies = map[string]string{}
 	for _, c := range pkg.Config.Cookies {
 		c = strings.TrimSuffix(c, "\r")
 		c = strings.TrimSpace(c)
@@ -250,11 +254,7 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 			cSlice := strings.SplitAfterN(c, "=", 2)
 			cSlice[0] = strings.TrimSuffix(cSlice[0], "=")
 
-			cookie := http.Cookie{
-				Name:  cSlice[0],
-				Value: cSlice[1],
-			}
-			pkg.Config.Website.Cookies = append(pkg.Config.Website.Cookies, &cookie)
+			pkg.Config.Website.Cookies[cSlice[0]] = cSlice[1]
 		}
 	}
 
@@ -272,8 +272,6 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 	http.DefaultClient = clientNoRedir
 
-	pkg.Statistics[stat+"total"]++
-
 	// retrieve cookies, headers etc. Only setStatusCode if no cookies shall be accepted. Otherwise the next request with set Cookies sets the status code
 	if !pkg.Config.DeclineCookies {
 		pkg.Config.Website, err = pkg.GetWebsite(u, false, false)
@@ -287,16 +285,12 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 		msg := fmt.Sprintf("Couldn't test url: %s\n", err.Error())
 		pkg.Print(msg, pkg.Red)
-
-		pkg.Statistics[stat+"noconnection"]++
 		return
 	}
 
 	if strings.HasPrefix(pkg.Config.Website.Body, "<html><head><title>Burp Suite") {
 		msg := fmt.Sprintf("Couldn't connect to given url: \n%s\n", pkg.Config.Website.Body)
 		pkg.Print(msg, pkg.Red)
-
-		pkg.Statistics[stat+"noconnection"]++
 		return
 	}
 
@@ -317,7 +311,7 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 	// check if there's a cache and the cachebuster works
 	var errSlice []error
 	var alwaysMiss bool
-	pkg.Config.Website.Cache, alwaysMiss, errSlice = pkg.CheckCache(stat, parameterList, headerList)
+	pkg.Config.Website.Cache, alwaysMiss, errSlice = pkg.CheckCache(parameterList, headerList)
 	for _, err := range errSlice {
 		if err != nil {
 			repWebsite.HasError = true
@@ -340,8 +334,6 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 			msg := fmt.Sprintf("Couldn't test url: %s\n", err.Error())
 			pkg.Print(msg, pkg.Red)
-
-			pkg.Statistics[stat+"errorcachebuster"]++
 			return
 		}
 	}
@@ -351,7 +343,6 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 		pkg.PrintVerbose(msg, pkg.NoColor, 1)
 
 		if alwaysMiss || pkg.Config.Website.Cache.Indicator == "" {
-			pkg.Statistics[stat+"deceptiontested"]++
 			repWebsite.Results = append(repWebsite.Results, pkg.TestWebCacheDeception())
 		} else {
 			msg = "The response already gets cached!"
@@ -364,6 +355,10 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 	/*******************************************/
 	if pkg.Config.Website.Cache.CBwasFound || pkg.Config.Force {
+		repWebsite.CacheIndicator = pkg.Config.Website.Cache.Indicator
+		repWebsite.CBName = pkg.Config.Website.Cache.CBName
+		repWebsite.CBwasFound = pkg.Config.Website.Cache.CBwasFound
+
 		if !pkg.Config.Website.Cache.CBwasFound && pkg.Config.Force {
 			fmt.Println()
 			pkg.Print("No Cachebuster was found. Forcing the parameter "+pkg.Config.CacheBuster+" as Cachebuster.\n", pkg.Cyan)
@@ -372,8 +367,6 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 			pkg.Config.Website.Cache.CBName = pkg.Config.CacheBuster
 		}
 
-		pkg.Statistics[stat+"poisoningtested"]++
-
 		/* Testing for cookie poisoning */
 		if noTestPreference || strings.Contains(pkg.Config.OnlyTest, "cookie") || (pkg.Config.SkipTest != "" && !strings.Contains(pkg.Config.SkipTest, "cookie")) {
 			msg = addSeparator("Cookie Poisoning")
@@ -509,6 +502,17 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 			pkg.PrintVerbose(msg, pkg.NoColor, 1)
 		}
 		/**********************************/
+
+		/* Testing for Parameter Encoding */
+		if noTestPreference || strings.Contains(pkg.Config.OnlyTest, "encoding") || (pkg.Config.SkipTest != "" && !strings.Contains(pkg.Config.SkipTest, "encoding")) {
+			msg = addSeparator("Parameter Encoding Poisoning")
+			pkg.PrintVerbose(msg, pkg.NoColor, 1)
+			repWebsite.Results = append(repWebsite.Results, pkg.ScanParameterEncoding())
+		} else {
+			msg = addSeparator("Skipping Parameter Encoding Poisoning")
+			pkg.PrintVerbose(msg, pkg.NoColor, 1)
+		}
+		/**********************************/
 	}
 
 	/* Check for linked files */
@@ -519,7 +523,7 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 		tempUrls := pkg.CrawlUrls(u, added, excluded)
 
 		if len(tempUrls) > 0 {
-			msg = fmt.Sprintf("Found %d urls", len(tempUrls))
+			msg = fmt.Sprintf("Found %d url(s)\n", len(tempUrls))
 			pkg.Print(msg, pkg.NoColor)
 
 			msg = "Adding the following urls to the Queue:"
@@ -530,7 +534,7 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 			*foundUrls = append(*foundUrls, tempUrls...)
 		} else {
-			msg = "No urls were found to add to the queue"
+			msg = "No urls were found to add to the queue\n"
 			pkg.Print(msg, pkg.NoColor)
 		}
 	}
@@ -554,7 +558,7 @@ func runTests(rec int, u string, progress string, foundUrls *[]string, stat stri
 
 		report.Vulnerable = report.Vulnerable || repWebsite.Vulnerable
 		pkg.PrintNewLine()
-		pkg.GenerateReport(report, currentDate)
+		pkg.GenerateReport(report, filePath)
 	}
 
 	pkg.Print("===============================================================\n\n", pkg.NoColor)
@@ -566,7 +570,7 @@ func addSeparator(msg string) string {
 }
 
 func createCompletedURLs() *os.File {
-	completedPath := pkg.Config.GeneratePath + currentDate + "_WCVS_Completed"
+	completedPath := filePath + "_Completed.txt"
 
 	_, err := os.Stat(completedPath)
 
@@ -587,50 +591,3 @@ func createCompletedURLs() *os.File {
 
 	return file
 }
-
-/* Setting proxy with specified proxyURL and proxyCertPath */
-func setProxy() *url.URL {
-	if pkg.Config.ProxyCertPath != "" {
-		proxyURL, err := url.Parse(pkg.Config.ProxyURL)
-		if err != nil {
-			msg := "setProxy: url.Parse: " + err.Error() + "\n"
-			pkg.PrintFatal(msg)
-		}
-		caCert, err := os.ReadFile(pkg.Config.ProxyCertPath)
-		if err != nil {
-			msg := "setProxy: ioutil.ReadFile: " + err.Error() + "\n"
-			pkg.PrintFatal(msg)
-		}
-		//caCertPool,err := x509.SystemCertPool()
-		// führt zu crypto/x509: system root pool is not available on Windows
-		caCertPool := x509.NewCertPool()
-
-		ok := caCertPool.AppendCertsFromPEM(caCert)
-		if !ok {
-			msg := "setProxy: could not append cert\n"
-			pkg.PrintFatal(msg)
-		}
-
-		tlsConfig := &tls.Config{RootCAs: caCertPool,
-			InsecureSkipVerify: true}
-		tr := &http.Transport{
-			Proxy:           http.ProxyURL(proxyURL),
-			TLSClientConfig: tlsConfig}
-
-		err = http2.ConfigureTransport(tr)
-		if err != nil {
-			msg := fmt.Sprintf("setProxy: Cannot switch to HTTP2: %s\n", err.Error())
-			pkg.PrintFatal(msg)
-		}
-
-		http.DefaultTransport = tr
-		return proxyURL
-	} else {
-		tlsConfig := &tls.Config{InsecureSkipVerify: true}
-		tr := &http.Transport{
-			TLSClientConfig: tlsConfig}
-
-		http.DefaultTransport = tr
-		return nil
-	}
-}
